/* $Id: cpio.c,v 1.8 2006/12/26 15:52:25 khorben Exp $ */



#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "plugin.h"


/* types */
#pragma pack(1)
struct cpio_ascii
{
	char magic[6];
	char dev[6];
	char ino[6];
	char mode[6];
	char uid[6];
	char gid[6];
	char nlink[6];
	char rdev[6];
	char mtime[11];
	char namesize[6];
	char filesize[11];
};

struct cpio_binary
{
	uint16_t magic;
	uint16_t dev;
	uint16_t ino;
	uint16_t mode;
	uint16_t uid;
	uint16_t gid;
	uint16_t nlink;
	uint16_t rdev;
	uint32_t mtime;
	uint16_t namesize;
	uint32_t filesize;
};
#pragma pack()

union cpio
{
	struct cpio_ascii ascii;
	struct cpio_binary binary;
};


/* variables */
/* signatures */
static char sig_ascii[] = "070707";
static char sig_binary_big[] = "\x71\xc7";
static char sig_binary_little[] = "\xc7\x71";
static char sig_trailer[] = "TRAILER!!";

/* plugin */
PluginType plugin_type = PT_ARCHIVE;
char const plugin_name[] = "CPIO";
PluginMagic plugin_magic[] =
{
	{ sizeof(struct cpio_ascii),	0, sig_ascii,	sizeof(sig_ascii)-1   },
	{ sizeof(struct cpio_binary),	0, sig_binary_big,
		sizeof(sig_binary_big)-1  },
	{ sizeof(struct cpio_binary),	0, sig_binary_little,
		sizeof(sig_binary_little)-1 },
	{ sizeof(union cpio),		0, sig_trailer,	sizeof(sig_trailer)-1 },
	{ 0,				0, NULL,	0		      }
};


/* functions */
static int _callback_ascii(PluginHelper * ph, struct cpio_ascii * cpio);
static int _callback_binary_big(PluginHelper * ph, struct cpio_binary * cpio);
static int _callback_binary_little(PluginHelper * ph,
		struct cpio_binary * cpio);
int plugin_callback(PluginHelper * ph, FILE * fp)
{
	fpos_t pos;
	union cpio buf;
	char trailer[sizeof(sig_trailer)];
	size_t len;

	if(fgetpos(fp, &pos) != 0)
		return -1;
	if(fread(&buf, sizeof(union cpio), 1, fp) == 1)
	{
		if(memcmp(&buf.ascii.magic, sig_ascii, sizeof(sig_ascii)-1)
				== 0)
			return _callback_ascii(ph, &buf.ascii);
		if(memcmp(&buf.binary.magic, sig_binary_big,
					sizeof(sig_binary_big)-1) == 0)
			return _callback_binary_big(ph, &buf.binary);
		if(memcmp(&buf.binary.magic, sig_binary_little,
					sizeof(sig_binary_little)-1) == 0)
			return _callback_binary_little(ph, &buf.binary);
	}
	if(fsetpos(fp, &pos) == 0
			&& fread(&buf, sizeof(struct cpio_ascii), 1, fp) == 1
			&& memcmp(&buf.ascii.magic, sig_ascii,
				sizeof(sig_ascii)-1) == 0)
		return _callback_ascii(ph, &buf.ascii);
	if(fsetpos(fp, &pos) == 0
			&& (len = fread(trailer, 1, sizeof(trailer)-1, fp)) > 0
			&& len == sizeof(trailer)-1
			&& strncmp(trailer, sig_trailer, len) == 0)
	{
		ph->printf(ph, "%s", "trailer\n");
		return 50;
	}
	return 0;
}

static int _callback_ids(PluginHelper * ph, uid_t uid, gid_t gid);
static int _callback_mtime(PluginHelper * ph, time_t mtime);
static int _callback_ascii(PluginHelper * ph, struct cpio_ascii * cpio)
{
	int score = 0;
	char buf[12];
	uid_t uid;
	gid_t gid;
	time_t mtime;
	char * p;

	ph->printf(ph, "%s", "ascii");
	memcpy(buf, cpio->uid, sizeof(cpio->uid));
	buf[sizeof(cpio->uid)] = '\0';
	uid = strtol(buf, &p, 8);
	if(*buf == '\0' || *p != '\0')
		ph->printf(ph, "%s", " uid invalid");
	else
	{
		memcpy(buf, cpio->gid, sizeof(cpio->gid));
		buf[sizeof(cpio->gid)] = '\0';
		gid = strtol(buf, &p, 8);
		if(*buf == '\0' || *p != '\0')
			ph->printf(ph, "%s", ", gid invalid");
		else
			score+=_callback_ids(ph, uid, gid) / 2;
	}
	memcpy(buf, cpio->mtime, sizeof(cpio->mtime));
	buf[sizeof(cpio->mtime)] = '\0';
	mtime = strtol(buf, &p, 8);
	if(*buf == '\0' || *p != '\0')
		ph->printf(ph, "%s", " invalid time");
	else
		score+=_callback_mtime(ph, mtime) / 2;
	ph->printf(ph, "\n");
	return score;
}

static int _callback_ids(PluginHelper * ph, uid_t uid, gid_t gid)
{
	ph->printf(ph, "%s%u%s%u", ", uid ", uid, ", gid ", gid);
	if(uid == gid)
		return 20;
	return 0;
}

static int _callback_mtime(PluginHelper * ph, time_t mtime)
{
	struct tm tm;
	char tmp[22] = "";

	if(mtime == 0)
		return 100;
	if((gmtime_r(&mtime, &tm)) == NULL || strftime(tmp, sizeof(tmp),
				", %d/%m/%Y %H:%M:%S", &tm) == 0)
	{
		ph->printf(ph, "%s", ", unknown date");
		return 0;
	}
	ph->printf(ph, "%s", tmp);
	/* between Wed May  1 00:00:00 CEST 1996 and now */
	return mtime > 830901600 && mtime < time(NULL) ? 100 : 0;
}

static int _binary_do(PluginHelper * ph, struct cpio_binary * cpio);
static int _callback_binary_big(PluginHelper * ph, struct cpio_binary * cpio)
{
	ph->printf(ph, "%s", "binary big-endian");
	cpio->uid = htob16(cpio->uid);
	cpio->gid = htob16(cpio->gid);
	cpio->mtime = htob32(cpio->mtime);
	return _binary_do(ph, cpio);
}

static int _binary_do(PluginHelper * ph, struct cpio_binary * cpio)
{
	int score = 0;

	score+=_callback_ids(ph, cpio->uid, cpio->gid) / 2;
	score+=_callback_mtime(ph, cpio->mtime) / 2;
	ph->printf(ph, "\n");
	return score;
}

static int _callback_binary_little(PluginHelper * ph, struct cpio_binary * cpio)
{
	ph->printf(ph, "%s", "binary little-endian");
	cpio->uid = htol16(cpio->uid);
	cpio->gid = htol16(cpio->gid);
	cpio->mtime = htol32(cpio->mtime);
	return _binary_do(ph, cpio);
}
